'''
Created on 17 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from shapely.geometry import LineString, Point, Polygon
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.RayCastingLib import RayCastingLib


class STIsovistField2D(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, buildings, viewpoints, nRays=64, rayLength=100.0):
        '''
        Constructor
        '''
        if not isinstance(buildings, GeoDataFrame):
            raise IllegalArgumentTypeException(buildings, 'buildings GeoDataFrame')
        if not isinstance(viewpoints, GeoDataFrame):
            raise IllegalArgumentTypeException(viewpoints, 'viewpoints GeoDataFrame')
        if not (buildings.crs == viewpoints.crs):
            raise Exception('Illegal argument: buildings and viewpoints must share shames CRS!')

        self.buildings = buildings
        self.masksSIdx = buildings.sindex
        self.viewpoints = viewpoints
        self.shootingDirs = RayCastingLib.preparePanopticRays(nRays)
        self.rayLength = rayLength

    def run(self):
        rowsIsovRays, rowsIsov = list(), list()
        for _, row in self.viewpoints.iterrows():
            vp = Point((row.geometry.x, row.geometry.y))  # To handle Point Z
            rays, ctrPoints, _, _ = RayCastingLib.multipleRayCast2D(self.buildings, self.masksSIdx,
                                                                    vp, self.shootingDirs, self.rayLength)

            isov = Polygon(ctrPoints) 
            isovCentre = isov.centroid
            drift = LineString([vp, isovCentre])

            rowsIsovRays.append(self.updateOrAppend(row, { 
                'geometry': rays, 'viewpoint': vp.wkt, 'vect_drift': drift.wkt }))
            rowsIsov.append(self.updateOrAppend(row, {
                'geometry': isov, 'viewpoint': vp.wkt, 'vect_drift': drift.wkt }))

        return [
            GeoDataFrame(rowsIsovRays, crs=self.viewpoints.crs),
            GeoDataFrame(rowsIsov, crs=self.viewpoints.crs)
            ]



'''
Created on 19 mars 2021

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
import geopandas
from geopandas.geodataframe import GeoDataFrame
from matplotlib.colors import ListedColormap
from shapely.affinity import scale, translate
from shapely.geometry.multipolygon import MultiPolygon
from shapely.ops import unary_union, triangulate
from shapely.wkt import loads
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.morph.STPointsDensifier import STPointsDensifier

import matplotlib.pyplot as plt 


class Logos(object):
    '''
    classdocs
    '''

    @staticmethod
    def logoAsMultiLineString(_deltaX=50.0, _scaleX=0.9):
        _t = loads('MULTILINESTRING ((0 100, 0 0, 50 0), (0 50, 50 50))')

        _4 = loads('LINESTRING (50 50, 0 50, 30 100, 30 0)')
        _4 = translate(_4, xoff=_deltaX)

        _g = loads('LINESTRING (50 0, 0 0, 0 50, 50 50, 50 -50, 0 -50)')
        _g = translate(_g, xoff=2 * _deltaX)

        _p = loads('LINESTRING (0 -50, 0 50, 50 50, 50 0, 0 0)')
        _p = translate(_p, xoff=3 * _deltaX)

        _d = loads('LINESTRING (50 50, 0 50, 0 0, 50 0, 50 100)')
        _d = translate(_d, xoff=4 * _deltaX)

        _t4gpd = [_t, _4, _g, _p, _d]
        _t4gpd = [scale(letter, xfact=_scaleX) for letter in _t4gpd]
        _t4gpd = unary_union(_t4gpd)

        return _t4gpd

    @staticmethod
    def logoAsMultiPolygon(_deltaX=50.0, _scaleX=0.75, _buffDist=5.0):
        _mls = Logos.logoAsMultiLineString(_deltaX, _scaleX)
        return _mls.buffer(_buffDist)

    @staticmethod
    def logoAsMesh():
        gdf = GeoDataFrame([{'geometry':Logos.logoAsMultiPolygon()}])

        _gdf = STPointsDensifier(gdf, distance=45.0, pathidFieldname=None, adjustableDist=True, removeDuplicate=True).run()
        _tin = triangulate(unary_union(GeomLib.getListOfShapelyPoints(unary_union(_gdf.geometry))))
        _gdf = GeoDataFrame([{'geometry': MultiPolygon(_tin)}]).explode()
        gdf.geometry = gdf.geometry.apply(lambda g: g.buffer(1e-3))

        _gdf = geopandas.overlay(_gdf, gdf, how='intersection')
        _gdf['gid'] = range(len(_gdf))
        _gdf['gid'] = _gdf.gid % 4

        return _gdf

    @staticmethod
    def logoAsPdf(filename=None, greyscaled=False):
        _gdf = Logos.logoAsMesh()

        fig, _basemap = plt.subplots(figsize=(1 * 8.26, 1 * 8.26))

        if greyscaled:
            my_rgb_cmap = ListedColormap(['lightgrey', 'grey', 'darkgrey', 'dimgrey', 'black'])
            my_rgb_cmap = ListedColormap(['grey', 'darkgrey', 'dimgrey'])
            _gdf.plot(ax=_basemap, column='gid', edgecolor='grey', linewidth=0.1, cmap=my_rgb_cmap)
        else:
            _gdf.plot(ax=_basemap, color='dimgrey', edgecolor='lightgrey', linewidth=0.1)

        _basemap.axis('off')
        if filename is None:
            plt.show()
        else:
            fig.savefig(filename, bbox_inches='tight')



'''
Created on 29 sept. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from t4gpd.commons.GeoProcess import GeoProcess
from geopandas.geodataframe import GeoDataFrame
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
import matplotlib.pyplot as plt


class FrequencyHistogram(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, inputGdf, fieldnames, nClusters, minValues=None, maxValues=None, outputFile='',
                 density=False, cumulative=False):
        '''
        Constructor
        '''
        if not isinstance(inputGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(inputGdf, 'GeoDataFrame')
        self.inputGdf = inputGdf

        if isinstance(fieldnames, str):
            self.fieldnames = [fieldnames]
        elif isinstance(fieldnames, (list, tuple)):
            self.fieldnames = fieldnames
        else:
            raise IllegalArgumentTypeException(fieldnames, 'must be a list or tuple of fieldnames')
        for fieldname in self.fieldnames:
            if fieldname not in self.inputGdf:
                raise Exception('%s is not a relevant field name!' % (fieldname))
        self.nFields = len(self.fieldnames)

        self.nClusters = nClusters if isinstance(nClusters, (list, tuple)) else [nClusters] * self.nFields

        if minValues is None:
            self.minValues = [None] * self.nFields
        else:
            self.minValues = minValues if isinstance(minValues, (list, tuple)) else [minValues] * self.nFields
        if maxValues is None:
            self.maxValues = [None] * self.nFields
        else:
            self.maxValues = maxValues if isinstance(maxValues, (list, tuple)) else [maxValues] * self.nFields

        if (self.nFields != len(self.nClusters)):
            raise IllegalArgumentTypeException(
                self.nClusters, 'nClusters must be either an int or a list of %d items' % self.nFields)
        if (self.nFields != len(self.minValues)):
            raise IllegalArgumentTypeException(
                self.minValues, 'minValues must be either a float or a list of %d items' % self.nFields)
        if (self.nFields != len(self.maxValues)):
            raise IllegalArgumentTypeException(
                self.maxValues, 'maxValues must be either a float or a list of %d items' % self.nFields)

        self.outputFile = outputFile
        self.density = density
        self.cumulative = cumulative

    def run(self):
        values = [list(self.inputGdf[fieldname]) for fieldname in self.fieldnames]

        numcols, numrows, pairsOfWeightsAndRanges = 1, self.nFields, []

        mainFig = plt.figure(figsize=(numcols * 8.26, numrows * 8.26))  # 21 cm ~ 8.26 inches
        mainFig.clear()

        for fignum in range(self.nFields):
            plt.subplot(numrows, numcols, fignum + 1)  # numrows, numcols, fignum     
            plt.title('Histogram of %s' % self.fieldnames[fignum])
            plt.xlabel(self.fieldnames[fignum])
            if self.density:
                plt.ylabel('Probability density')
            else:
                plt.ylabel('Number of values')
            plt.grid(True)

            n, bins, _ = plt.hist(values[fignum], bins=self.nClusters[fignum],
                                  density=self.density, cumulative=self.cumulative,
                                  range=(self.minValues[fignum], self.maxValues[fignum]),
                                  facecolor='gray', edgecolor='black', orientation='vertical',
                                  hatch='', histtype='bar', align='mid', rwidth=1.0)
            x = [(bins[i] + bins[i + 1]) / 2.0 for i in range(len(bins) - 1)]
            pairsOfWeightsAndRanges.append([n, bins])
            plt.plot(x, n, 'ro', x, n, 'r--')

        if self.outputFile is not None:
            if ('' == self.outputFile):
                plt.show()
            else:
                plt.savefig(self.outputFile, format='pdf', bbox_inches='tight')  # , facecolor='w', edgecolor='k', pad_inches=0.15)

        return pairsOfWeightsAndRanges



'''
Created on 11 sept. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from shapely.affinity import rotate
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess


class Rotation2D(AbstractGeoprocess):
    '''
    classdocs
    '''

    def __init__(self, angle, origin='center', use_radians=False):
        '''
        Constructor
        '''
        self.angle = angle
        self.origin = origin
        self.use_radians = use_radians

    def runWithArgs(self, row):
        return { 'geometry': rotate(
            row.geometry, self.angle, origin='center', use_radians=False) }



'''
Created on 25 aug. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from datetime import datetime

from t4gpd.commons.AngleLib import AngleLib

import matplotlib.pyplot as plt
from t4gpd.energy.Dogniaux import Dogniaux
from t4gpd.energy.Perez import Perez
from t4gpd.energy.Perraudeau import Perraudeau
from t4gpd.energy.PerrinDeBrichambaut import PerrinDeBrichambaut


class PlotDirectNormalIrradiance(object):
    '''
    classdocs
    '''
    @staticmethod
    def plot(dt=datetime(2020, 3, 21)):
        # PEREZ MODEL PARAMETERS:
        delta = 0.12
        epsilon = 6.3
        dayInYear = dt.timetuple().tm_yday

        mainFig = plt.figure(1)
        mainFig.clear()

        x = list(range(0, 91, 1))

        y1 = [Dogniaux.directNormalIrradiance(AngleLib.toRadians(_x)) for _x in x]
        y2 = [Perez.directSolarIrradiance(AngleLib.toRadians(_x), dayInYear, delta, epsilon) for _x in x]
        y3 = [Perraudeau.directNormalIrradiance(AngleLib.toRadians(_x)) for _x in x]
        y4 = [PerrinDeBrichambaut.directNormalIrradiance(AngleLib.toRadians(_x), skyType=PerrinDeBrichambaut.PURE_SKY) for _x in x]
        y5 = [PerrinDeBrichambaut.directNormalIrradiance(AngleLib.toRadians(_x), skyType=PerrinDeBrichambaut.STANDARD_SKY) for _x in x]
        y6 = [PerrinDeBrichambaut.directNormalIrradiance(AngleLib.toRadians(_x), skyType=PerrinDeBrichambaut.POLLUTED_SKY) for _x in x]

        plt.title('Direct Normal Irradiance')
        plt.xlabel('Solar Altitude Angle')
        plt.ylabel('Direct Normal Irradiance (W/m2)')
        plt.grid(True)
        _d, _p1, _p2, _pdb_c, _pdb_s, _pdb_p = plt.plot(x, y1, 'b-', x, y2, 'c-', x, y3, 'k-', x, y4, 'm--', x, y5, 'g--', x, y6, 'r--')
        plt.legend((_d, _p1, _p2, _pdb_c, _pdb_s, _pdb_p), (
            'Dogniaux\'s model',
            'Perez\'s model',
            'Perraudeau\'s model',
            'Perrin de Brichambaut\'s model (Pure sky)',
            'Perrin de Brichambaut\'s model (Standard sky)',
            'Perrin de Brichambaut\'s model (Polluted sky)',
        ))
        plt.show()

# PlotDirectNormalIrradiance.plot()



'''
Created on 18 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame

import geopandas as gpd
from t4gpd.commons.BoundingBox import BoundingBox
from t4gpd.commons.GeoProcess import GeoProcess


class STLoadAndClip(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, filename, roi):
        '''
        Constructor
        '''
        self.inputGdf = gpd.read_file(filename)
        self.spatialIdx = self.inputGdf.sindex
        self.roi = BoundingBox(roi).asPolygon()

    def run(self):
        rows = []

        ids = list(self.spatialIdx.intersection(self.roi.bounds))
        for _id in ids:
            row = self.inputGdf.loc[_id]
            rowGeom = row.geometry
            if self.roi.intersects(rowGeom):
                _row = { 'geometry':rowGeom.intersection(self.roi) }
                rows.append(self.updateOrAppend(row, _row))
        return GeoDataFrame(rows, crs=self.inputGdf.crs)



'''
Created on 17 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from shapely.geometry import Point

from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.MyEdge import MyEdge
from t4gpd.commons.MyNode import MyNode


class GeoWriter(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, inputGdf, outputFile, characteristicLength=10.0, toLocalCrs=True):
        '''
        Constructor
        '''
        if not isinstance(inputGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(inputGdf, 'GeoDataFrame')
        self.inputGdf = inputGdf

        self.outputFile = outputFile
        self.lc = characteristicLength
        self.toLocalCrs = toLocalCrs

    def run(self):
        newOrigin = Point((0.0, 0.0, 0.0))
        if self.toLocalCrs:
            minx, miny, _, _ = self.inputGdf.total_bounds
            newOrigin = Point((minx, miny, 0.0))

        mapOfNodes = {}
        mapOfEdges = {}
        arrayOfLineLoops = []
        arrayOfPlaneSurfaces = []

        for _, row in self.inputGdf.iterrows():
            geom = row.geometry
            if GeomLib.isMultipart(geom):
                for g in geom.geoms:
                    self.__dumpPolygon(g, mapOfNodes, mapOfEdges, arrayOfLineLoops, arrayOfPlaneSurfaces)
            else:
                self.__dumpPolygon(geom, mapOfNodes, mapOfEdges, arrayOfLineLoops, arrayOfPlaneSurfaces)
 
        with open(self.outputFile, 'w') as output: 
            self.__dumpNodes(output, mapOfNodes, newOrigin)
            self.__dumpEdges(output, mapOfEdges, mapOfNodes)
            self.__dumpLineLoops(output, arrayOfLineLoops)
            self.__dumpPlaneSurfaces(output, arrayOfPlaneSurfaces)

    def __dumpPolygon(self, polygon, mapOfNodes, mapOfEdges, arrayOfLineLoops, arrayOfPlaneSurfaces):
        tmp = list()
        tmp.append(self.__fulfillWith(mapOfNodes, mapOfEdges, arrayOfLineLoops, polygon.exterior))
        for hole in polygon.interiors:
            tmp.append(self.__fulfillWith(mapOfNodes, mapOfEdges, arrayOfLineLoops, hole))
        arrayOfPlaneSurfaces.append(tmp)
        
    def __fulfillWith(self, mapOfNodes, mapOfEdges, arrayOfLineLoops, ring):
        nodes, tmp = list(ring.coords)[:-1], list()

        prev = MyNode(nodes[-1])
        for curr in nodes:
            curr = MyNode(curr)
            if curr not in mapOfNodes:
                mapOfNodes[curr] = 1 + len(mapOfNodes)

            myEdge = MyEdge(prev, curr)
            if myEdge not in mapOfEdges:
                mapOfEdges[myEdge] = 1 + len(mapOfEdges)
            myEdgeIdx = mapOfEdges[myEdge]

            if myEdge.startNode.__eq__(prev):
                tmp.append(myEdgeIdx)
            else:
                tmp.append(-myEdgeIdx)

            prev = curr

        arrayOfLineLoops.append(tmp)
        return len(arrayOfLineLoops)

    def __dumpCoords(self, node, newOrigin):
        return '%.2f, %.2f, %.2f' % (node.x - newOrigin.x, node.y - newOrigin.y, node.z - newOrigin.z)

    def __dumpNodes(self, output, mapOfNodes, newOrigin):
        output.write('lc = %g;\n' % self.lc)
        output.write('\n// =====> %d POINT(S)\n' % len(mapOfNodes))

        # for key, value in sorted(mapOfNodes.iteritems(), key=lambda(k, v): (v)):
        for key, value in sorted(list(mapOfNodes.items()), key=lambda kv: kv[1]):
            output.write('Point(%d) = {%s, lc};\n' % (value, self.__dumpCoords(key, newOrigin)))

    def __dumpEdges(self, output, mapOfEdges, mapOfNodes):
        output.write('\n// =====> %d LINE(S)\n' % len(mapOfEdges))
        # for key, value in sorted(mapOfEdges.iteritems(), key=lambda(k, v): v):
        for key, value in sorted(list(mapOfEdges.items()), key=lambda kv: kv[1]):
            output.write('Line(%d) = {%d, %d};\n' % (value, mapOfNodes[key.startNode], mapOfNodes[key.endNode]))

    def __dumpLineLoops(self, output, arrayOfLineLoops):
        output.write('\n// =====> %d LINE LOOP(S)\n' % len(arrayOfLineLoops))
        for i in range(len(arrayOfLineLoops)):
            output.write('Line Loop(%d) = {%s};\n' % (i + 1, ', '.join([str(x) for x in arrayOfLineLoops[i]])))

    def __dumpPlaneSurfaces(self, output, arrayOfPlaneSurfaces):
        output.write('\n// =====> %d PLANE SURFACE(S)\n' % len(arrayOfPlaneSurfaces))
        for i in range(len(arrayOfPlaneSurfaces)):
            output.write('Plane Surface(%d) = {%s};\n' % (i + 1, ', '.join([str(x) for x in arrayOfPlaneSurfaces[i]])))
